#!/usr/bin/env python

"""
charting for spec2nexus

.. autosummary::

    ~make_png
    ~xy_plot

"""

import datetime
import numpy
import matplotlib
import matplotlib.figure
matplotlib.use('Agg')
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

import spec2nexus


SCALING_FACTOR = 1        #  2**24
PLOT_H_INT = 9          # 7
PLOT_V_INT = 5          # 3
COLORMAP = 'cubehelix'        # http://matplotlib.org/api/pyplot_summary.html#matplotlib.pyplot.colormaps
WATERMARK_TEXT = '%s, (C) %s' % (spec2nexus.__package_name__, spec2nexus.__copyright__.split(',')[0])


def make_png(
        image, 
        image_file, 
        axes = None,
        title = '2-D data',
        subtitle = '',
        log_image=False, 
        hsize=PLOT_H_INT, 
        vsize=PLOT_V_INT, 
        cmap=COLORMAP,
        xtitle=None, ytitle=None, 
        timestamp_str=None):
    """
    read the image from the named HDF5 file and make a PNG file
    
    Test that the HDF5 file exists and that the path to the data exists in that file.
    Read the data from the named dataset, mask off some bad values, 
    convert to log(image) and use Matplotlib to make the PNG file.
    
    :param obj image: array of data to be rendered
    :param str image_file: name of image file to be written (path is optional)
    :param bool log_image: plot log(image)
    :param int hsize: horizontal size of the PNG image (default: 7)
    :param int hsize: vertical size of the PNG image (default: 3)
    :param str cmap: colormap for the image (default: 'cubehelix'), 'jet' is another good one
    :return str: *image_file*
    
    The HDF5 file could be a NeXus file, or some other layout.
    """

    # replace masked data with min good value
    image_data = numpy.ma.masked_less_equal(image, 0)
    image_data = image_data.filled(image_data.min())
    image_data = numpy.ma.masked_invalid(image_data)    # mask out any NaN values

    if log_image and image_data.max() != 0:     # apply log scaling
        image_data = numpy.log(image_data)
        image_data -= image_data.min()
        image_data *= SCALING_FACTOR / image_data.max()

    fig = matplotlib.figure.Figure(figsize=(hsize, vsize))
    fig.clf()
    ax = fig.add_subplot('111')
    if isinstance(axes, list) and len(axes) == 2:
        def shift_to_pixel_boundaries(arr):
            np_arr = numpy.array(arr)
            step = np_arr[1] - np_arr[0]
            return numpy.append(np_arr, np_arr[-1] + step) - step/2
        
        x = shift_to_pixel_boundaries(axes[0])
        y = shift_to_pixel_boundaries(axes[1])

        ax.pcolor(x, y, image_data, cmap=cmap)
        ax.axis([x.min(), x.max(), y.min(), y.max()])
    else:
        ax.imshow(image_data, interpolation='nearest', cmap=cmap)

    if xtitle is not None:
        ax.set_xlabel(xtitle)
    if ytitle is not None:
        ax.set_ylabel(ytitle)

    timestamp_str = timestamp_str or str(datetime.datetime.now())
    
    if subtitle is not None:
        ax.set_title(subtitle, fontsize=10)
    fig.suptitle(title, fontsize=8)
    fig.text(0.02, 0., timestamp_str,
        fontsize=8, color='gray',
        ha='left', va='bottom', alpha=0.5)
    fig.text(0.98, 0., WATERMARK_TEXT,
        fontsize=8, color='gray',
        ha='right', va='bottom', alpha=0.5)

    ax.ticklabel_format(useOffset=False, style='plain')

    FigureCanvas(fig).print_figure(image_file, bbox_inches='tight')

    return image_file


def xy_plot(
        x, y, 
        plot_file, 
        title=None, subtitle=None, 
        xtitle=None, ytitle=None, 
        xlog=False, ylog=False,
        hsize=PLOT_H_INT, 
        vsize=PLOT_V_INT, 
        timestamp_str=None):
    r"""
    with MatPlotLib, generate a plot of a scan (as if data from a scan in a SPEC file)
    
    :param [float] x: horizontal axis data
    :param [float] y: vertical axis data
    :param str plot_file: file name to write plot image
    :param str xtitle: horizontal axis label (default: not shown)
    :param str ytitle: vertical axis label (default: not shown)
    :param str title: title for plot (default: date time)
    :param str subtitle: subtitle for plot (default: not shown)
    :param bool xlog: should X axis be log (default: False=linear)
    :param bool ylog: should Y axis be log (default: False=linear)
    :param str timestamp_str: date to use on plot (default: now)

    .. tip:: when using this module as a background task ...
    
        MatPlotLib has several interfaces for plotting. 
        Since this module runs as part of a background job 
        generating lots of plots, MatPlotLib's standard ``plt`` code is 
        not the right model.  It warns after 20 plots and 
        will eventually run out of memory.  
        
        Here's the fix used in this module:
        http://stackoverflow.com/questions/16334588/create-a-figure-that-is-reference-counted/16337909#16337909

    """
    fig = matplotlib.figure.Figure(figsize=(hsize, vsize))
    fig.clf()

    ax = fig.add_subplot('111')
    if xlog:
        ax.set_xscale('log')
        if max(x) <= 0:
            msg = 'X data has no positive values,'
            msg += ' and therefore can not be log-scaled.'
            raise ValueError(msg)
    if ylog:
        ax.set_yscale('log')
        if max(y) <= 0:
            msg = 'Y data has no positive values,'
            msg += ' and therefore can not be log-scaled.'
            raise ValueError(msg)
    if not xlog and not ylog:
        ax.ticklabel_format(useOffset=False)
    if xtitle is not None:
        ax.set_xlabel(xtitle)
    if ytitle is not None:
        ax.set_ylabel(ytitle)

    if subtitle is not None:
        ax.set_title(subtitle, fontsize=10)
    fig.suptitle(title, fontsize=8)
    fig.text(0.02, 0., timestamp_str,
        fontsize=8, color='gray',
        ha='left', va='bottom', alpha=0.5)
    fig.text(0.98, 0., WATERMARK_TEXT,
        fontsize=8, color='gray',
        ha='right', va='bottom', alpha=0.5)

    if not xlog and not ylog:
        ax.ticklabel_format(useOffset=False, style='plain')
    ax.plot(x, y, 'o-')

    FigureCanvas(fig).print_figure(plot_file, bbox_inches='tight')



#!/usr/bin/env python 
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
**#H** & **#V** - Metadata in SPEC data files as defined by APS UNICAT

Handles the UNICAT control lines which write additional metadata
in the scans using #H/#V pairs of labels/values.
"""


import re
import six

from .. import eznx
from ..plugin import AutoRegister, ControlLineHandler
from ..utils import strip_first_word


@six.add_metaclass(AutoRegister)
class UNICAT_MetadataMnemonics(ControlLineHandler):

    """
    **#H** -- UNICAT metadata names (numbered rows: #H0, #H1, ...)
    
    Individual metadata names are expected to be single-word strings 
    but may be multi-word strings as long as the words in the string 
    are separated by only one space.  The delimiter between metadata 
    names is two consecutive spaces.  A tab (``\\t``) character is also 
    acceptable but should be avoided.

    IN-MEMORY REPRESENTATION
    
    * (SpecDataFileHeader) : **H** : labels
    * (SpecDataFileScan): **metadata** : {labels: values}
    
    HDF5/NeXus REPRESENTATION
    
    * *NXnote* group named **metadata** below the 
      *NXentry* group, such as **/entry/metadata**
      
      * datasets created from dictionary <scan>.metadata

    """

    key = '#H\d+'
    scan_attributes_defined = ['H']
    
    def process(self, text, spec_obj, *args, **kws):
        row_text = strip_first_word(text)
        labels = re.split(r"  +", row_text)
        spec_obj.H.append(labels)


@six.add_metaclass(AutoRegister)
class UNICAT_MetadataValues(ControlLineHandler):

    """
    **#V** -- UNICAT metadata values (numbered rows: #V0, #V1, ...)
    
    Individual metadata values are expected to be numbers but may 
    be multi-word strings as long as the words in the string are 
    separated by only one space.  The delimiter between metadata 
    values is two consecutive spaces.  A tab (``'\\t\``) character is 
    also acceptable but should be avoided.
    
    All numerical values will be converted into floating point
    numbers.  Only if that conversion fails, the text of the value 
    will be reported verbatim.

    IN-MEMORY REPRESENTATION
    
    * (SpecDataFileScan): **V** : values
    * (SpecDataFileScan): **metadata** : {labels: values}
    
    HDF5/NeXus REPRESENTATION
    
    * *NXnote* group named **metadata** below the 
      *NXentry* group, such as **/entry/metadata**
      
      * datasets created from dictionary <scan>.metadata

    """

    key = '#V\d+'
    scan_attributes_defined = ['V', 'metadata']
    
    def process(self, text, scan, *args, **kws):
        index = len(scan.V)
        row_text =  strip_first_word(text)
        for delimiter in (r" "*2, r" "):
            values = re.split(delimiter, row_text)
            if len(scan.header.H[index]) == len(values):
                    break

        scan.V.append(values)
        scan.addPostProcessor('unicat_metadata', self.postprocess)
    
    def postprocess(self, scan, *args, **kws):
        """
        interpret the UNICAT metadata (mostly floating point) from the scan header
        
        :param SpecDataFileScan scan: data from a single SPEC scan (instance of SpecDataFileScan)
        """
        scan.metadata = {}
        if not hasattr(scan.header, "H"):
            msg = "No matching #H line(s) for scan %d" % scan.scanNum
            raise KeyError(msg)
        for row, values in enumerate(scan.V):
            if (row+1) > len(scan.header.H):
                msg = "No matching #H%d line for #V%d in scan %d" % (row, row, scan.scanNum)
                raise KeyError(msg)
            for col, val in enumerate(values):
                if (col+1) > len(scan.header.H[row]):
                    msg = "No matching label in #H%d line for #V%d, column %d in scan %d" % (row, row, col, scan.scanNum)
                    raise KeyError(msg)
                label = scan.header.H[row][col]
                try:
                    scan.metadata[label] = float(val)
                except ValueError:
                    scan.metadata[label] = val
        scan.addH5writer(self.key, self.writer)
    
    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """Describe how to store this data in an HDF5 NeXus file"""
        if hasattr(scan, 'metadata') and len(scan.metadata) > 0:
            desc='SPEC metadata (UNICAT-style #H & #V lines)'
            group = eznx.makeGroup(h5parent, 'metadata', nxclass, description=desc)
            writer.save_dict(group, scan.metadata)

            # link it to the NXinstrument group
            nxinstrument = eznx.openGroup(h5parent, 'instrument', "NXinstrument")
            eznx.makeLink(h5parent, group, nxinstrument.name + "/metadata")



#!/usr/bin/env python 
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
Fallback handling for any SPEC data file control lines not recognized by other handlers
"""

from collections import OrderedDict
import six

from ..eznx import makeGroup
from ..plugin import AutoRegister, ControlLineHandler
from ..spec import UNRECOGNIZED_KEY, SpecDataFileHeader, SpecDataFileScan


@six.add_metaclass(AutoRegister)
class UnrecognizedControlLine(ControlLineHandler):
    
    """unrecognized control line"""

    key = UNRECOGNIZED_KEY
    scan_attributes_defined = ['_unrecognized']

    def process(self, text, spec_obj, *args, **kws):
        ' '
        if not hasattr(spec_obj, '_unrecognized'):
            spec_obj._unrecognized = []
        spec_obj._unrecognized.append(text)
        if isinstance(spec_obj, SpecDataFileHeader) or isinstance(spec_obj, SpecDataFileScan):
            spec_obj.addH5writer(self.key, self.writer)
    
    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """write the data in a NeXus group named ``unrecognized``"""
        desc = "SPEC data file control lines not otherwise recognized"
        nxclass = 'NXnote'
        i = 0
        success = False
        while not success:
            i += 1
            nm = 'unrecognized_' + str(i)
            if nm not in h5parent.keys():
                group = makeGroup(h5parent, nm, nxclass, description=desc)
                success = True
        dd = OrderedDict()
        for i, value in enumerate(scan._unrecognized):
            dd['u' + str(i)] = value
        writer.save_dict(group, dd)




"""
singletons: Python 2 and 3 Compatible Version

:see: http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

USAGE::

    class Logger(Singleton):
        pass

"""

class _Singleton(type):

    """A metaclass that creates a Singleton base class when called."""

    _instances = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._instances[self] = super(_Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]

class Singleton(_Singleton('SingletonMeta', (object,), {})):
    """Public interface"""




#!/usr/bin/python

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------


"""
Save columns from SPEC data file scan(s) to TSV files

.. note:: TSV: tab-separated values

**Usage**::

  extractSpecScan.py /tmp/CeCoIn5 -s 5 -c HerixE Ana5 ICO-C
  extractSpecScan.py ./testdata/11_03_Vinod.dat   -s 2 12   -c USAXS.m2rp Monitor  I0

.. note:: sdpecified column names **MUST** appear in all chosen scans

Compatible with Python 2.7+
"""

from __future__ import print_function

__url__ = 'http://spec2nexus.readthedocs.org/en/latest/extractSpecScan.html'

import os
import sys
from . import spec


#-------------------------------------------------------------------------------------------


REPORTING_QUIET    = 'quiet'
REPORTING_STANDARD = 'standard'
REPORTING_VERBOSE  = 'verbose'


#-------------------------------------------------------------------------------------------


def makeOutputFileName(specFile, scanNum):
    """
    return an output file name based on specFile and scanNum
    
    :param str specFile: name of existing SPEC data file to be read
    :param str scanNum: number of chosen SPEC scan

    append scanNum to specFile to get output file name 
    (before file extension if present)
    
    Always add a file extension to the output file.
    If none is present, use ".dat".
    
    Examples:
    
    ===========  ========   ===============
    specFile     scanNum    outFile
    ===========  ========   ===============
    CeCoIn5      scan 5     CeCoIn5_5.dat
    CeCoIn5.dat  scan 77    CeCoIn5_77.dat
    CeCoIn5.dat  scan 5.1   CeCoIn5_5_1.dat
    ===========  ========   ===============
    """
    name_parts = os.path.splitext(specFile)
    scan_number = str(scanNum).replace('.', '_')
    outFile = name_parts[0] + '_' + scan_number + name_parts[1]
    return outFile


def expand_scan_range_terms(scans):
    # expand range of scan terms into individual scan numbers
    _scans_ = []
    for item in scans:
        if item.startswith('-'):
            _scans_.append(item)
        else:
            parts = item.split('-')
            if len(parts) == 1:
                _scans_.append(item)
            elif len(parts) == 2:
                _scans_ += [str(k) for k in range(int(parts[0]), int(parts[1])+1)]
            else:
                raise ValueError('scan specifier not recognized: ' + item)
    return _scans_


def get_user_parameters():
    """configure user's command line parameters from sys.argv"""
    import argparse
    from spec2nexus._version import get_versions
    version = get_versions()['version']

    doc = __doc__.strip().splitlines()[0]
    doc += '\n  URL: ' + __url__
    doc += '\n  v' + version
    parser = argparse.ArgumentParser(prog='extractSpecScan', description=doc)

    parser.add_argument('-v',
                        '--version', 
                        action='version',
                        help='print version number and exit',
                        version=version)

    msg = 'do not write column labels to output file (default: write labels)'
    parser.add_argument('--nolabels', 
                        action='store_true',
                        help=msg,
                        default=False)

    parser.add_argument('spec_file',
                        action='store', 
                        help="SPEC data file name(s)")

    msg = "scan number(s) to be extracted (must specify at least one) a range is allowed: 8-12 (no whitespace around the '-')"
    parser.add_argument('-s',
                        '--scan', 
                        action='store', 
                        nargs='+', 
                        #type=int,
                        required=True,
                        help=msg)
    
    msg = "column label(s) to be extracted (must specify at least one)"
    parser.add_argument('-c',
                        '--column', 
                        action='store',
                        nargs='+', 
                        required=True,
                        help=msg)
    
    msg = "report scan Geometry (#G) header information"
    parser.add_argument('-G',
                        action='store_true',
                        default=False,
                        help=msg)
    
    msg = "report scan Positioners (#O & #P) header information"
    parser.add_argument('-P',
                        action='store_true',
                        default=False,
                        help=msg)
    
    msg = "report scan Q (#Q) header information"
    parser.add_argument('-Q',
                        action='store_true',
                        default=False,
                        help=msg)
    
    msg = "report scan (UNICAT-style #H & #V) header information"
    parser.add_argument('-V',
                        action='store_true',
                        default=False,
                        help=msg)

    group = parser.add_mutually_exclusive_group()
    group.set_defaults(reporting_level=REPORTING_STANDARD)
    
    msg =  'suppress all program output (except errors)'
    msg += ', do not use with --verbose option'
    group.add_argument('--quiet', 
                       dest='reporting_level',
                       action='store_const',
                       const=REPORTING_QUIET,
                       help=msg)
    
    msg =  'print more program output'
    msg += ', do not use with --quiet option'
    group.add_argument('--verbose', 
                       dest='reporting_level',
                       action='store_const',
                       const=REPORTING_VERBOSE,
                       help=msg)

    args = parser.parse_args()
    
    args.scan = expand_scan_range_terms(args.scan)
    
    args.print_labels = not args.nolabels
    del args.nolabels

    return args


def main():
    """
    read the data file, find each scan, find the columns, save the data
    
    :param [str] cmdArgs: Namespace from argparse, returned from get_user_parameters()
    
    ..  such as:
      Namespace(column=['mr', 'I0', 'USAXS_PD'], print_labels=True, scan=[1, 6], spec_file=['data/APS_spec_data.dat'])
    
    .. note:: Each column label must match *exactly* the name of a label
       in each chosen SPEC scan number or the program will skip that particular scan
       
       If more than one column matches, the first match will be selected.
    
    example output::

        # mr    I0    USAXS_PD
        1.9475    65024    276
        1.9725    64845    352
        1.9975    65449    478
    
    """
    cmdArgs = get_user_parameters()

    if cmdArgs.reporting_level in (REPORTING_STANDARD, REPORTING_VERBOSE):
        print("program: " + sys.argv[0])
    # now open the file and read it
    specData = spec.SpecDataFile(cmdArgs.spec_file)
    if cmdArgs.reporting_level in (REPORTING_STANDARD, REPORTING_VERBOSE):
        print("read: " + cmdArgs.spec_file)
    
    for scanNum in cmdArgs.scan:
        outFile = makeOutputFileName(cmdArgs.spec_file, scanNum)
        scan = specData.getScan(scanNum)
        scan.interpret()       # force the plug-ins to be processed
    
        # get the column numbers corresponding to the column_labels
        column_numbers = []
        for label in cmdArgs.column:
            if label in scan.L:
                # report all columns in order specified on command-line
                column_numbers.append( scan.L.index(label) )
            else:
                if cmdArgs.reporting_level in (REPORTING_VERBOSE):
                    msg = 'column label "' + label + '" not found in scan #'
                    msg += str(scanNum) + ' ... skipping'
                    print(msg)       # report all mismatched column labels
    
        if len(column_numbers) == len(cmdArgs.column):   # must be perfect matches
            txt = []
            if cmdArgs.print_labels:
                txt.append( '# ' + '\t'.join(cmdArgs.column) )
            if len(scan.data) > 0:
                data = [scan.data[item] for item in cmdArgs.column]
                for data_row in zip(*data):
                    txt.append( '\t'.join(map(str, data_row)) )
            else:
                # issue #45: no data lines in scan
                txt.append( '# <no SPEC data points in this scan>' )
                
            header_data = []
            if cmdArgs.G:
                for k, v in sorted(scan.G.items()):
                    # use tab separation to make it easy to pull into spreadsheet columns
                    header_data.append('#G\t%s\t%s' % (k, '\t'.join(v.split(' '))))
            if cmdArgs.P:
                for k, v in sorted(scan.positioner.items()):
                    header_data.append('#P\t%s\t%s' % (k, str(v)))  # str(v) vs. %g or other
            if cmdArgs.Q:
                if len(scan.Q) > 0:
                    header_data.append('#Q\t%s' % '\t'.join(map(str, scan.Q)))
            if cmdArgs.V:
                try:
                    for k, v in sorted(scan.metadata.items()):
                        header_data.append('#V\t%s\t%s' % (k, v))
                except AttributeError:
                    pass    # no metadata
            if len(header_data):
                header_data.insert(0, '# data from scan heading')
                header_data.insert(0, '#')
                header_data.append('#')
                header_data.append('# scan data:')
                header_data.append('#')

            # at the top of every file now
            header_data.insert(0, '# scan: ' + str(scanNum))
            header_data.insert(0, '# file: ' + str(cmdArgs.spec_file))
        
            fp = open(outFile, 'w')
            if len(header_data):
                fp.write('\n'.join(header_data) + '\n')
            fp.write('\n'.join(txt))
            fp.close()
            if cmdArgs.reporting_level in (REPORTING_STANDARD, REPORTING_VERBOSE):
                print("wrote: " + outFile)
            if cmdArgs.reporting_level in (REPORTING_VERBOSE,):
                for item in scan.get_interpreter_comments():
                    print(item)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE, distributed with this software.
#-----------------------------------------------------------------------------

from setuptools import setup
import os
import sys

import versioneer

# pull in some definitions from the package's __init__.py file
basedir = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(basedir, 'src', ))
import spec2nexus as package


verbose=1
long_description = open(os.path.join(basedir, 'README.md'), 'r').read()


setup (
    name             = package.__package_name__,
    version          = versioneer.get_version(),
    cmdclass         = versioneer.get_cmdclass(),
    license          = package.__license__,
    description      = package.__description__,
    long_description = long_description,
    author           = package.__author_name__,
    author_email     = package.__author_email__,
    url              = package.__url__,
    #download_url=package.__download_url__,
    keywords         = package.__keywords__,
    platforms        = 'any',
    install_requires = package.__install_requires__,
    package_dir      = {'': 'src'},
    packages         = ['spec2nexus', 'spec2nexus.plugins', ],
    package_data     = {
         'spec2nexus': [
            'data/02_03_setup_fly/*',
            'data/*.*',
            'data/Cd*',
            'plugins/*.xsd',
            'LICENSE.txt',
            'diffractometer-geometries.dict',
            ],
         },
    classifiers      = package.__classifiers__,
    entry_points     = {
         # create & install scripts in <python>/bin
         'console_scripts': [
             'spec2nexus=spec2nexus.nexus:main',
             'extractSpecScan=spec2nexus.extractSpecScan:main',
             'specplot=spec2nexus.specplot:main',
             'specplot_gallery=spec2nexus.specplot_gallery:main',
    	],
         #'gui_scripts': [],
    },
)
